import { StateValue, ActivityMap, EventObject, HistoryValue, ActionObject, EventType, StateConfig, SCXML, StateSchema, TransitionDefinition, Typestate } from './types';
import { StateNode } from './StateNode';
import { Actor } from './Actor';
export declare function stateValuesEqual(a: StateValue | undefined, b: StateValue | undefined): boolean;
export declare function isState<TContext, TEvent extends EventObject, TStateSchema extends StateSchema<TContext> = any, TTypestate extends Typestate<TContext> = any>(state: object | string): state is State<TContext, TEvent, TStateSchema, TTypestate>;
export declare function bindActionToState<TC, TE extends EventObject>(action: ActionObject<TC, TE>, state: State<TC, TE>): ActionObject<TC, TE>;
export declare class State<TContext, TEvent extends EventObject = EventObject, TStateSchema extends StateSchema<TContext> = any, TTypestate extends Typestate<TContext> = {
    value: any;
    context: TContext;
}> {
    value: StateValue;
    context: TContext;
    historyValue?: HistoryValue | undefined;
    history?: State<TContext, TEvent, TStateSchema>;
    actions: Array<ActionObject<TContext, TEvent>>;
    activities: ActivityMap;
    meta: any;
    events: TEvent[];
    event: TEvent;
    _event: SCXML.Event<TEvent>;
    _sessionid: string | null;
    /**
     * Indicates whether the state has changed from the previous state. A state is considered "changed" if:
     *
     * - Its value is not equal to its previous value, or:
     * - It has any new actions (side-effects) to execute.
     *
     * An initial state (with no history) will return `undefined`.
     */
    changed: boolean | undefined;
    /**
     * Indicates whether the state is a final state.
     */
    done: boolean | undefined;
    /**
     * The enabled state nodes representative of the state value.
     */
    configuration: Array<StateNode<TContext, any, TEvent>>;
    /**
     * The next events that will cause a transition from the current state.
     */
    nextEvents: EventType[];
    /**
     * The transition definitions that resulted in this state.
     */
    transitions: Array<TransitionDefinition<TContext, TEvent>>;
    /**
     * An object mapping actor IDs to spawned actors/invoked services.
     */
    children: Record<string, Actor>;
    /**
     * Creates a new State instance for the given `stateValue` and `context`.
     * @param stateValue
     * @param context
     */
    static from<TC, TE extends EventObject = EventObject>(stateValue: State<TC, TE> | StateValue, context?: TC | undefined): State<TC, TE>;
    /**
     * Creates a new State instance for the given `config`.
     * @param config The state config
     */
    static create<TC, TE extends EventObject = EventObject>(config: StateConfig<TC, TE>): State<TC, TE>;
    /**
     * Creates a new `State` instance for the given `stateValue` and `context` with no actions (side-effects).
     * @param stateValue
     * @param context
     */
    static inert<TC, TE extends EventObject = EventObject>(stateValue: State<TC, TE> | StateValue, context: TC): State<TC, TE>;
    /**
     * Creates a new State instance.
     * @param value The state value
     * @param context The extended state
     * @param historyValue The tree representing historical values of the state nodes
     * @param history The previous state
     * @param actions An array of action objects to execute as side-effects
     * @param activities A mapping of activities and whether they are started (`true`) or stopped (`false`).
     * @param meta
     * @param events Internal event queue. Should be empty with run-to-completion semantics.
     * @param configuration
     */
    constructor(config: StateConfig<TContext, TEvent>);
    /**
     * Returns an array of all the string leaf state node paths.
     * @param stateValue
     * @param delimiter The character(s) that separate each subpath in the string state node path.
     */
    toStrings(stateValue?: StateValue, delimiter?: string): string[];
    toJSON(): Pick<this, Exclude<keyof this, "configuration" | "transitions">>;
    /**
     * Whether the current state value is a subset of the given parent state value.
     * @param parentStateValue
     */
    matches<TSV extends TTypestate['value']>(parentStateValue: TSV): this is State<(TTypestate extends {
        value: TSV;
    } ? TTypestate : never)['context'], TEvent, TStateSchema, TTypestate> & {
        value: TSV;
    };
}
//# sourceMappingURL=State.d.ts.map